From 2755471f335624ad64751dc101aa23787e436be4 Mon Sep 17 00:00:00 2001
From: Thomas B. <thomas@roundcube.net>
Date: Fri, 06 Mar 2015 13:06:34 -0500
Subject: [PATCH] Merge pull request #257 from corbosman/session_refactor

---
 program/lib/Roundcube/rcube.php                  |   13 
 program/lib/Roundcube/rcube_session.php          |  414 +++++++---------------
 program/lib/Roundcube/rcube_session_memcache.php |  144 ++++++++
 program/lib/Roundcube/rcube_session_php.php      |   77 ++++
 program/lib/Roundcube/rcube_session_db.php       |  173 +++++++++
 program/lib/Roundcube/rcube_session_redis.php    |  211 +++++++++++
 config/defaults.inc.php                          |   14 
 7 files changed, 759 insertions(+), 287 deletions(-)

diff --git a/program/lib/Roundcube/rcube.php b/program/lib/Roundcube/rcube.php
index 20f509e..3d08153 100644
--- a/program/lib/Roundcube/rcube.php
+++ b/program/lib/Roundcube/rcube.php
@@ -521,23 +521,14 @@
         ini_set('session.use_only_cookies', 1);
         ini_set('session.cookie_httponly', 1);
 
-        // use database for storing session data
-        $this->session = new rcube_session($this->get_dbh(), $this->config);
-
-        $this->session->register_gc_handler(array($this, 'gc'));
-        $this->session->set_secret($this->config->get('des_key') . dirname($_SERVER['SCRIPT_NAME']));
-        $this->session->set_ip_check($this->config->get('ip_check'));
-
-        if ($this->config->get('session_auth_name')) {
-            $this->session->set_cookiename($this->config->get('session_auth_name'));
-        }
+        // get session driver instance
+        $this->session = rcube_session::factory($this->config);
 
         // start PHP session (if not in CLI mode)
         if ($_SERVER['REMOTE_ADDR']) {
             $this->session->start();
         }
     }
-
 
     /**
      * Garbage collector - cache/temp cleaner
diff --git a/program/lib/Roundcube/rcube_session.php b/program/lib/Roundcube/rcube_session.php
index 8306a06..fc1d871 100644
--- a/program/lib/Roundcube/rcube_session.php
+++ b/program/lib/Roundcube/rcube_session.php
@@ -15,37 +15,36 @@
  +-----------------------------------------------------------------------+
  | Author: Thomas Bruederli <roundcube@gmail.com>                        |
  | Author: Aleksander Machniak <alec@alec.pl>                            |
+ | Author: Cor Bosman <cor@roundcu.be>                            |
  +-----------------------------------------------------------------------+
 */
 
 /**
- * Class to provide database supported session storage
+ * Abstract class to provide database supported session storage
  *
  * @package    Framework
  * @subpackage Core
  * @author     Thomas Bruederli <roundcube@gmail.com>
  * @author     Aleksander Machniak <alec@alec.pl>
  */
-class rcube_session
+abstract class rcube_session
 {
-    private $db;
-    private $ip;
-    private $start;
-    private $changed;
-    private $time_diff = 0;
-    private $reloaded = false;
-    private $appends = array();
-    private $unsets = array();
-    private $gc_handlers = array();
-    private $cookiename = 'roundcube_sessauth';
-    private $vars;
-    private $key;
-    private $now;
-    private $secret = '';
-    private $ip_check = false;
-    private $logging = false;
-    private $storage;
-    private $memcache;
+    protected $key;
+    protected $ip;
+    protected $changed;
+    protected $start;
+    protected $time_diff = 0;
+    protected $reloaded = false;
+    protected $appends = array();
+    protected $unsets = array();
+    protected $gc_handlers = array();
+    protected $cookiename = 'roundcube_sessauth';
+    protected $vars;
+    protected $now;
+    protected $secret = '';
+    protected $ip_check = false;
+    protected $logging = false;
+    protected $config;
 
     /**
      * Blocks session data from being written to database.
@@ -54,58 +53,70 @@
      */
     public $nowrite = false;
 
+    /**
+     * Factory, returns driver-specific instance of the class
+     *
+     * @param object $config
+     * @return Object rcube_session
+     */
+    public static function factory($config)
+    {
+        // get session storage driver
+        $storage = $config->get('session_storage', 'db');
+
+        // class name for this storage
+        $class = "rcube_session_" . $storage;
+
+        // try to instantiate class
+        if (class_exists($class)) {
+            return new $class($config);
+        }
+
+        // no storage found, raise error
+        rcube::raise_error(array('code' => 604, 'type' => 'session',
+                               'line' => __LINE__, 'file' => __FILE__,
+                               'message' => "Failed to find session driver. Check session_storage config option"),
+                           true, true);
+    }
 
     /**
-     * Default constructor
+     * @param Object $config
      */
-    public function __construct($db, $config)
+    public function __construct($config)
     {
-        $this->db      = $db;
-        $this->start   = microtime(true);
-        $this->ip      = rcube_utils::remote_addr();
-        $this->logging = $config->get('log_session', false);
+        $this->config = $config;
 
-        $lifetime = $config->get('session_lifetime', 1) * 60;
-        $this->set_lifetime($lifetime);
+        // register default gc handler
+        $this->register_gc_handler(array($this, 'gc'));
 
-        // use memcache backend
-        $this->storage = $config->get('session_storage', 'db');
-        if ($this->storage == 'memcache') {
-            $this->memcache = rcube::get_instance()->get_memcache();
+        // set secret
+        $this->set_secret($this->config->get('des_key') . dirname($_SERVER['SCRIPT_NAME']));
 
-            // set custom functions for PHP session management if memcache is available
-            if ($this->memcache) {
-                ini_set('session.serialize_handler', 'php');
+        // set ip check
+        $this->set_ip_check($this->config->get('ip_check'));
 
-                session_set_save_handler(
-                    array($this, 'open'),
-                    array($this, 'close'),
-                    array($this, 'mc_read'),
-                    array($this, 'mc_write'),
-                    array($this, 'mc_destroy'),
-                    array($this, 'gc'));
-            }
-            else {
-                rcube::raise_error(array('code' => 604, 'type' => 'db',
-                    'line' => __LINE__, 'file' => __FILE__,
-                    'message' => "Failed to connect to memcached. Please check configuration"),
-                true, true);
-            }
+        // set cookie name
+        if ($this->config->get('session_auth_name')) {
+            $this->set_cookiename($this->config->get('session_auth_name'));
         }
-        else if ($this->storage != 'php') {
-            ini_set('session.serialize_handler', 'php');
+    }
 
-            // set custom functions for PHP session management
-            session_set_save_handler(
-                array($this, 'open'),
-                array($this, 'close'),
-                array($this, 'db_read'),
-                array($this, 'db_write'),
-                array($this, 'db_destroy'),
-                array($this, 'gc'));
+    /**
+     * register session handler
+     */
+    public function register_session_handler()
+    {
+        ini_set('session.serialize_handler', 'php');
 
-            $this->table_name = $this->db->table_name('session', true);
-        }
+        // set custom functions for PHP session management
+        session_set_save_handler(
+            array($this, 'open'),
+            array($this, 'close'),
+            array($this, 'read'),
+            array($this, 'sess_write'),
+            array($this, 'destroy'),
+            array($this, 'gc')
+        );
     }
 
 
@@ -114,37 +125,51 @@
      */
     public function start()
     {
+        $this->start   = microtime(true);
+        $this->ip      = rcube_utils::remote_addr();
+        $this->logging = $this->config->get('log_session', false);
+
+        $lifetime = $this->config->get('session_lifetime', 1) * 60;
+        $this->set_lifetime($lifetime);
+
         session_start();
-
-        // copy some session properties to object vars
-        if ($this->storage == 'php') {
-            $this->key     = session_id();
-            $this->ip      = $_SESSION['__IP'];
-            $this->changed = $_SESSION['__MTIME'];
-        }
     }
 
-
-    public function open($save_path, $session_name)
-    {
-        return true;
-    }
-
-
-    public function close()
-    {
-        return true;
-    }
+    /**
+     * Abstract methods should be implemented by driver classes
+     */
+    abstract function open($save_path, $session_name);
+    abstract function close();
+    abstract function destroy($key);
+    abstract function read($key);
+    abstract function write($key, $vars);
+    abstract function update($key, $newvars, $oldvars);
 
 
     /**
-     * Delete session data for the given key
+     * session write handler. This calls the implementation methods for write/update after some initial checks.
      *
-     * @param string Session ID
+     * @param $key
+     * @param $vars
+     * @return bool
      */
-    public function destroy($key)
+    public function sess_write($key, $vars)
     {
-        return $this->memcache ? $this->mc_destroy($key) : $this->db_destroy($key);
+        if ($this->nowrite) {
+            return true;
+        }
+
+        // check cache
+        $oldvars = $this->get_cache($key);
+
+        // if there are cached vars, update store, else insert new data
+        if ($oldvars !== null) {
+            $newvars = $this->_fixvars($vars, $oldvars);
+            return $this->update($key, $newvars, $oldvars);
+        }
+        else {
+            return $this->write($key, $vars);
+        }
     }
 
 
@@ -153,11 +178,6 @@
      */
     public function write_close()
     {
-        if ($this->storage == 'php') {
-            $_SESSION['__IP'] = $this->ip;
-            $_SESSION['__MTIME'] = time();
-        }
-
         session_write_close();
 
         // write_close() is called on script shutdown, see rcube::shutdown()
@@ -166,91 +186,10 @@
         $this->gc_shutdown();
     }
 
-
-    /**
-     * Read session data from database
-     *
-     * @param string Session ID
-     *
-     * @return string Session vars
-     */
-    public function db_read($key)
-    {
-        $sql_result = $this->db->query(
-            "SELECT `vars`, `ip`, `changed`, " . $this->db->now() . " AS ts"
-            . " FROM {$this->table_name} WHERE `sess_id` = ?", $key);
-
-        if ($sql_result && ($sql_arr = $this->db->fetch_assoc($sql_result))) {
-            $this->time_diff = time() - strtotime($sql_arr['ts']);
-            $this->changed   = strtotime($sql_arr['changed']);
-            $this->ip        = $sql_arr['ip'];
-            $this->vars      = base64_decode($sql_arr['vars']);
-            $this->key       = $key;
-
-            return !empty($this->vars) ? (string) $this->vars : '';
-        }
-
-        return null;
-    }
-
-
-    /**
-     * Save session data.
-     * handler for session_read()
-     *
-     * @param string Session ID
-     * @param string Serialized session vars
-     *
-     * @return boolean True on success
-     */
-    public function db_write($key, $vars)
-    {
-        $now = $this->db->now();
-        $ts  = microtime(true);
-
-        if ($this->nowrite)
-            return true;
-
-        // no session row in DB (db_read() returns false)
-        if (!$this->key) {
-            $oldvars = null;
-        }
-        // use internal data from read() for fast requests (up to 0.5 sec.)
-        else if ($key == $this->key && (!$this->vars || $ts - $this->start < 0.5)) {
-            $oldvars = $this->vars;
-        }
-        else { // else read data again from DB
-            $oldvars = $this->db_read($key);
-        }
-
-        if ($oldvars !== null) {
-            $newvars = $this->_fixvars($vars, $oldvars);
-
-            if ($newvars !== $oldvars) {
-                $this->db->query("UPDATE {$this->table_name} "
-                    . "SET `changed` = $now, `vars` = ? WHERE `sess_id` = ?",
-                    base64_encode($newvars), $key);
-            }
-            else if ($ts - $this->changed + $this->time_diff > $this->lifetime / 2) {
-                $this->db->query("UPDATE {$this->table_name} SET `changed` = $now"
-                    . " WHERE `sess_id` = ?", $key);
-            }
-        }
-        else {
-            $this->db->query("INSERT INTO {$this->table_name}"
-                . " (`sess_id`, `vars`, `ip`, `created`, `changed`)"
-                . " VALUES (?, ?, ?, $now, $now)",
-                $key, base64_encode($vars), (string)$this->ip);
-        }
-
-        return true;
-    }
-
-
     /**
      * Merge vars with old vars and apply unsets
      */
-    private function _fixvars($vars, $oldvars)
+    protected function _fixvars($vars, $oldvars)
     {
         if ($oldvars !== null) {
             $a_oldvars = $this->unserialize($oldvars);
@@ -280,97 +219,6 @@
         return $newvars;
     }
 
-
-    /**
-     * Handler for session_destroy()
-     *
-     * @param string Session ID
-     *
-     * @return boolean True on success
-     */
-    public function db_destroy($key)
-    {
-        if ($key) {
-            $this->db->query("DELETE FROM {$this->table_name} WHERE `sess_id` = ?", $key);
-        }
-
-        return true;
-    }
-
-
-    /**
-     * Read session data from memcache
-     *
-     * @param string Session ID
-     * @return string Session vars
-     */
-    public function mc_read($key)
-    {
-        if ($value = $this->memcache->get($key)) {
-            $arr = unserialize($value);
-            $this->changed = $arr['changed'];
-            $this->ip      = $arr['ip'];
-            $this->vars    = $arr['vars'];
-            $this->key     = $key;
-
-            return !empty($this->vars) ? (string) $this->vars : '';
-        }
-
-        return null;
-    }
-
-
-    /**
-     * Save session data.
-     * handler for session_read()
-     *
-     * @param string Session ID
-     * @param string Serialized session vars
-     *
-     * @return boolean True on success
-     */
-    public function mc_write($key, $vars)
-    {
-        $ts = microtime(true);
-
-        // no session data in cache (mc_read() returns false)
-        if (!$this->key)
-            $oldvars = null;
-        // use internal data for fast requests (up to 0.5 sec.)
-        else if ($key == $this->key && (!$this->vars || $ts - $this->start < 0.5))
-            $oldvars = $this->vars;
-        else // else read data again
-            $oldvars = $this->mc_read($key);
-
-        $newvars = $oldvars !== null ? $this->_fixvars($vars, $oldvars) : $vars;
-
-        if ($newvars !== $oldvars || $ts - $this->changed > $this->lifetime / 3) {
-            return $this->memcache->set($key, serialize(array('changed' => time(), 'ip' => $this->ip, 'vars' => $newvars)),
-                MEMCACHE_COMPRESSED, $this->lifetime + 60);
-        }
-
-        return true;
-    }
-
-
-    /**
-     * Handler for session_destroy() with memcache backend
-     *
-     * @param string Session ID
-     *
-     * @return boolean True on success
-     */
-    public function mc_destroy($key)
-    {
-        if ($key) {
-            // #1488592: use 2nd argument
-            $this->memcache->delete($key, 0);
-        }
-
-        return true;
-    }
-
-
     /**
      * Execute registered garbage collector routines
      */
@@ -380,7 +228,6 @@
         // see rcube::shutdown() and rcube_session::write_close()
         return $this->gc_enabled = $maxlifetime;
     }
-
 
     /**
      * Register additional garbage collector functions
@@ -405,12 +252,6 @@
     protected function gc_shutdown()
     {
         if ($this->gc_enabled) {
-            // just delete all expired sessions
-            if ($this->storage == 'db') {
-                $this->db->query("DELETE FROM {$this->table_name}"
-                    . " WHERE `changed` < " . $this->db->now(-$this->gc_enabled));
-            }
-
             foreach ($this->gc_handlers as $fct) {
                 call_user_func($fct);
             }
@@ -422,6 +263,7 @@
      * Generate and set new session id
      *
      * @param boolean $destroy If enabled the current session will be destroyed
+     * @return bool
      */
     public function regenerate_id($destroy=true)
     {
@@ -431,6 +273,28 @@
         $this->key  = session_id();
 
         return true;
+    }
+
+    /**
+     * see if we have vars of this key already cached, and if so, return them.
+     *
+     * @param $key
+     * @return null|array
+     */
+    protected function get_cache($key)
+    {
+        // no session data in cache (read() returns false)
+        if (!$this->key) {
+            $cache = null;
+        }
+        // use internal data for fast requests (up to 0.5 sec.)
+        else if ($key == $this->key && (!$this->vars || $ts - $this->start < 0.5)) {
+            $cache = $this->vars;
+        }
+        else { // else read data again
+            $cache = $this->read($key);
+        }
+        return $cache;
     }
 
 
@@ -523,10 +387,9 @@
             $node[$k] = $value;
         }
 
-        if ($this->key && $this->memcache)
-            $data = $this->mc_read($this->key);
-        else if ($this->key)
-            $data = $this->db_read($this->key);
+        if($this->key) {
+            $data = $this->read($this->key);
+        }
 
         if ($data) {
             session_decode($data);
@@ -553,7 +416,7 @@
      * Returns a reference to the node in data array referenced by the given path.
      * e.g. ['compose','attachments'] will return $_SESSION['compose']['attachments']
      */
-    private function &get_node($path, &$data_arr)
+    protected function &get_node($path, &$data_arr)
     {
         $node = &$data_arr;
         if (!empty($path)) {
@@ -570,7 +433,7 @@
     /**
      * Serialize session data
      */
-    private function serialize($vars)
+    protected function serialize($vars)
     {
         $data = '';
         if (is_array($vars)) {
@@ -589,7 +452,7 @@
      * Unserialize session data
      * http://www.php.net/manual/en/function.session-decode.php#56106
      */
-    private function unserialize($str)
+    protected function unserialize($str)
     {
         $str    = (string)$str;
         $endptr = strlen($str);
@@ -788,6 +651,7 @@
      * Create session cookie from session data
      *
      * @param int Time slot to use
+     * @return string
      */
     function _mkcookie($timeslot)
     {
diff --git a/program/lib/Roundcube/rcube_session_db.php b/program/lib/Roundcube/rcube_session_db.php
new file mode 100644
index 0000000..78138d1
--- /dev/null
+++ b/program/lib/Roundcube/rcube_session_db.php
@@ -0,0 +1,173 @@
+<?php
+
+/*
+ +-----------------------------------------------------------------------+
+ | This file is part of the Roundcube Webmail client                     |
+ | Copyright (C) 2005-2014, The Roundcube Dev Team                       |
+ | Copyright (C) 2011, Kolab Systems AG                                  |
+ |                                                                       |
+ | Licensed under the GNU General Public License version 3 or            |
+ | any later version with exceptions for skins & plugins.                |
+ | See the README file for a full license statement.                     |
+ |                                                                       |
+ | PURPOSE:                                                              |
+ |   Provide database supported session management                       |
+ +-----------------------------------------------------------------------+
+ | Author: Thomas Bruederli <roundcube@gmail.com>                        |
+ | Author: Aleksander Machniak <alec@alec.pl>                            |
+ | Author: Cor Bosman <cor@roundcu.be>                                   |
+ +-----------------------------------------------------------------------+
+*/
+
+/**
+ * Class to provide database session storage
+ *
+ * @package    Framework
+ * @subpackage Core
+ * @author     Thomas Bruederli <roundcube@gmail.com>
+ * @author     Aleksander Machniak <alec@alec.pl>
+ * @author     Cor Bosman <cor@roundcu.be>
+ */
+class rcube_session_db extends rcube_session
+{
+    private $db;
+    private $table_name;
+
+    /**
+     * @param Object $config
+     */
+    public function __construct($config)
+    {
+        parent::__construct($config);
+
+        // get db instance
+        $this->db = rcube::get_instance()->get_dbh();
+
+        // session table name
+        $this->table_name = $this->db->table_name('session', true);
+
+        // register sessions handler
+        $this->register_session_handler();
+
+        // register db gc handler
+        $this->register_gc_handler(array($this, 'gc_db'));
+    }
+
+    /**
+     * @param $save_path
+     * @param $session_name
+     * @return bool
+     */
+    public function open($save_path, $session_name)
+    {
+        return true;
+    }
+
+    /**
+     * @return bool
+     */
+    public function close()
+    {
+        return true;
+    }
+
+
+    /**
+     * Handler for session_destroy()
+     *
+     * @param $key
+     * @return bool
+     */
+    public function destroy($key)
+    {
+        if ($key) {
+            $this->db->query("DELETE FROM {$this->table_name} WHERE `sess_id` = ?", $key);
+        }
+
+        return true;
+    }
+
+    /**
+     * Read session data from database
+     *
+     * @param string Session ID
+     *
+     * @return string Session vars
+     */
+    public function read($key)
+    {
+        $sql_result = $this->db->query(
+            "SELECT `vars`, `ip`, `changed`, " . $this->db->now() . " AS ts"
+            . " FROM {$this->table_name} WHERE `sess_id` = ?", $key);
+
+        if ($sql_result && ($sql_arr = $this->db->fetch_assoc($sql_result))) {
+            $this->time_diff = time() - strtotime($sql_arr['ts']);
+            $this->changed   = strtotime($sql_arr['changed']);
+            $this->ip        = $sql_arr['ip'];
+            $this->vars      = base64_decode($sql_arr['vars']);
+            $this->key       = $key;
+
+            return !empty($this->vars) ? (string) $this->vars : '';
+        }
+        return null;
+    }
+
+    /**
+     * insert new data into db session store
+     *
+     * @param $key
+     * @param $vars
+     * @return bool
+     */
+    public function write($key, $vars)
+    {
+        $now        = $this->db->now();
+
+        $this->db->query("INSERT INTO {$this->table_name}"
+                         . " (`sess_id`, `vars`, `ip`, `created`, `changed`)"
+                         . " VALUES (?, ?, ?, $now, $now)",
+                         $key, base64_encode($vars), (string)$this->ip);
+
+        return true;
+    }
+
+
+    /**
+     * update session data
+     *
+     * @param $key
+     * @param $newvars
+     * @param $oldvars
+     *
+     * @return bool
+     */
+    public function update($key, $newvars, $oldvars)
+    {
+        $now        = $this->db->now();
+
+        // if new and old data are not the same, update data
+        // else update expire timestamp only when certain conditions are met
+        if ($newvars !== $oldvars) {
+            $this->db->query("UPDATE {$this->table_name} "
+                             . "SET `changed` = $now, `vars` = ? WHERE `sess_id` = ?",
+                             base64_encode($newvars), $key);
+        }
+        else if ($ts - $this->changed + $this->time_diff > $this->lifetime / 2) {
+            $this->db->query("UPDATE {$this->table_name} SET `changed` = $now"
+                                 . " WHERE `sess_id` = ?", $key);
+        }
+
+        return true;
+    }
+
+    /**
+     * Clean up db sessions.
+     */
+    public function gc_db()
+    {
+        // just clean all old sessions when this GC is called
+        $this->db->query("DELETE FROM " . $this->db->table_name('session')
+                         . " WHERE changed < " . $this->db->now(-$this->gc_enabled));
+    }
+
+}
\ No newline at end of file
diff --git a/program/lib/Roundcube/rcube_session_memcache.php b/program/lib/Roundcube/rcube_session_memcache.php
new file mode 100644
index 0000000..732d5fb
--- /dev/null
+++ b/program/lib/Roundcube/rcube_session_memcache.php
@@ -0,0 +1,144 @@
+<?php
+
+/*
+ +-----------------------------------------------------------------------+
+ | This file is part of the Roundcube Webmail client                     |
+ | Copyright (C) 2005-2014, The Roundcube Dev Team                       |
+ | Copyright (C) 2011, Kolab Systems AG                                  |
+ |                                                                       |
+ | Licensed under the GNU General Public License version 3 or            |
+ | any later version with exceptions for skins & plugins.                |
+ | See the README file for a full license statement.                     |
+ |                                                                       |
+ | PURPOSE:                                                              |
+ |   Provide database supported session management                       |
+ +-----------------------------------------------------------------------+
+ | Author: Thomas Bruederli <roundcube@gmail.com>                        |
+ | Author: Aleksander Machniak <alec@alec.pl>                            |
+ | Author: Cor Bosman <cor@roundcu.bet>                                   |
+ +-----------------------------------------------------------------------+
+*/
+
+/**
+ * Class to provide memcache session storage
+ *
+ * @package    Framework
+ * @subpackage Core
+ * @author     Thomas Bruederli <roundcube@gmail.com>
+ * @author     Aleksander Machniak <alec@alec.pl>
+ * @author     Cor Bosman <cor@roundcu.be>
+ */
+class rcube_session_memcache extends rcube_session
+{
+    private $memcache;
+
+    /**
+     * @param Object $config
+     */
+    public function __construct($config)
+    {
+        parent::__construct($config);
+
+        $this->memcache = rcube::get_instance()->get_memcache();
+
+        if (!$this->memcache) {
+            rcube::raise_error(array('code' => 604, 'type' => 'db',
+                                   'line' => __LINE__, 'file' => __FILE__,
+                                   'message' => "Failed to connect to memcached. Please check configuration"),
+                               true, true);
+        }
+
+        // register sessions handler
+        $this->register_session_handler();
+    }
+
+    /**
+     * @param $save_path
+     * @param $session_name
+     * @return bool
+     */
+    public function open($save_path, $session_name)
+    {
+        return true;
+    }
+
+    /**
+     * @return bool
+     */
+    public function close()
+    {
+        return true;
+    }
+
+    /**
+     * Handler for session_destroy() with memcache backend
+     *
+     * @param $key
+     * @return bool
+     */
+    public function destroy($key)
+    {
+        if ($key) {
+            // #1488592: use 2nd argument
+            $this->memcache->delete($key, 0);
+        }
+
+        return true;
+    }
+
+
+    /**
+     * Read session data from memcache
+     *
+     * @param $key
+     * @return null|string
+     */
+    public function read($key)
+    {
+        if ($value = $this->memcache->get($key)) {
+            $arr = unserialize($value);
+            $this->changed = $arr['changed'];
+            $this->ip      = $arr['ip'];
+            $this->vars    = $arr['vars'];
+            $this->key     = $key;
+
+            return !empty($this->vars) ? (string) $this->vars : '';
+        }
+
+        return null;
+    }
+
+    /**
+     * write data to memcache storage
+     *
+     * @param $key
+     * @param $vars
+     * @return bool
+     */
+    public function write($key, $vars)
+    {
+        return $this->memcache->set($key, serialize(array('changed' => time(), 'ip' => $this->ip, 'vars' => $vars)),
+                                    MEMCACHE_COMPRESSED, $this->lifetime + 60);
+    }
+
+    /**
+     * update memcache session data
+     *
+     * @param $key
+     * @param $newvars
+     * @param $oldvars
+     * @return bool
+     */
+    public function update($key, $newvars, $oldvars)
+    {
+        $ts = microtime(true);
+
+        if ($newvars !== $oldvars || $ts - $this->changed > $this->lifetime / 3) {
+            return $this->memcache->set($key, serialize(array('changed' => time(), 'ip' => $this->ip, 'vars' => $newvars)),
+                                        MEMCACHE_COMPRESSED, $this->lifetime + 60);
+        }
+
+        return true;
+    }
+
+}
\ No newline at end of file
diff --git a/program/lib/Roundcube/rcube_session_php.php b/program/lib/Roundcube/rcube_session_php.php
new file mode 100644
index 0000000..2f7085f
--- /dev/null
+++ b/program/lib/Roundcube/rcube_session_php.php
@@ -0,0 +1,77 @@
+<?php
+
+/*
+ +-----------------------------------------------------------------------+
+ | This file is part of the Roundcube Webmail client                     |
+ | Copyright (C) 2005-2014, The Roundcube Dev Team                       |
+ | Copyright (C) 2011, Kolab Systems AG                                  |
+ |                                                                       |
+ | Licensed under the GNU General Public License version 3 or            |
+ | any later version with exceptions for skins & plugins.                |
+ | See the README file for a full license statement.                     |
+ |                                                                       |
+ | PURPOSE:                                                              |
+ |   Provide database supported session management                       |
+ +-----------------------------------------------------------------------+
+ | Author: Thomas Bruederli <roundcube@gmail.com>                        |
+ | Author: Aleksander Machniak <alec@alec.pl>                            |
+ | Author: Cor Bosman <cor@roundcu.be>                                   |
+ +-----------------------------------------------------------------------+
+*/
+
+/**
+ * Class to provide native php session storage
+ *
+ * @package    Framework
+ * @subpackage Core
+ * @author     Thomas Bruederli <roundcube@gmail.com>
+ * @author     Aleksander Machniak <alec@alec.pl>
+ * @author     Cor Bosman <cor@roundcu.be>
+ */
+class rcube_session_php extends rcube_session {
+
+    /**
+     * native php sessions don't need a save handler
+     * we do need to define abstract function implementations but they are not used.
+     */
+
+    public function open($save_path, $session_name) {}
+    public function close() {}
+    public function destroy($key) {}
+    public function read($key) {}
+    public function write($key, $vars) {}
+    public function update($key, $newvars, $oldvars) {}
+
+    /**
+     * @param Object $config
+     */
+    public function __construct($config)
+    {
+        parent::__construct($config);
+    }
+
+    /**
+     * Wrapper for session_write_close()
+     */
+    public function write_close()
+    {
+        $_SESSION['__IP'] = $this->ip;
+        $_SESSION['__MTIME'] = time();
+
+        parent::write_close();
+    }
+
+    /**
+     * Wrapper for session_start()
+     */
+    public function start()
+    {
+        parent::start();
+
+        $this->key     = session_id();
+        $this->ip      = $_SESSION['__IP'];
+        $this->changed = $_SESSION['__MTIME'];
+
+    }
+
+}
\ No newline at end of file
diff --git a/program/lib/Roundcube/rcube_session_redis.php b/program/lib/Roundcube/rcube_session_redis.php
new file mode 100644
index 0000000..4822db7
--- /dev/null
+++ b/program/lib/Roundcube/rcube_session_redis.php
@@ -0,0 +1,211 @@
+<?php
+
+/*
+ +-----------------------------------------------------------------------+
+ | This file is part of the Roundcube Webmail client                     |
+ | Copyright (C) 2005-2014, The Roundcube Dev Team                       |
+ |                                                                       |
+ | Licensed under the GNU General Public License version 3 or            |
+ | any later version with exceptions for skins & plugins.                |
+ | See the README file for a full license statement.                     |
+ |                                                                       |
+ | PURPOSE:                                                              |
+ |   Provide redis supported session management                       |
+ +-----------------------------------------------------------------------+
+ | Author: Cor Bosman <cor@roundcu.be>                                   |
+ +-----------------------------------------------------------------------+
+*/
+
+/**
+ * Class to provide redis session storage
+ *
+ * @package    Framework
+ * @subpackage Core
+ * @author     Cor Bosman <cor@roundcu.be>
+ */
+class rcube_session_redis extends rcube_session {
+
+    private $redis;
+
+    /**
+     * @param Object $config
+     */
+    public function __construct($config)
+    {
+        parent::__construct($config);
+
+        // instantiate Redis object
+        $this->redis = new Redis();
+
+        if (!$this->redis) {
+            rcube::raise_error(array('code' => 604, 'type' => 'session',
+                                   'line' => __LINE__, 'file' => __FILE__,
+                                   'message' => "Failed to find Redis. Make sure php-redis is included"),
+                               true, true);
+        }
+
+        // get config instance
+        $hosts = $this->config->get('redis_hosts', array('localhost'));
+
+        // host config is wrong
+        if (!is_array($hosts) || empty($hosts)) {
+            rcube::raise_error(array('code' => 604, 'type' => 'session',
+                                   'line' => __LINE__, 'file' => __FILE__,
+                                   'message' => "Redis host not configured"),
+                               true, true);
+        }
+
+        // only allow 1 host for now until we support clustering
+        if (count($hosts) > 1) {
+            rcube::raise_error(array('code' => 604, 'type' => 'session',
+                                   'line' => __LINE__, 'file' => __FILE__,
+                                   'message' => "Redis cluster not yet supported"),
+                               true, true);
+        }
+
+        foreach ($hosts as $host) {
+            // explode individual fields
+            list($host, $port, $database, $password) = array_pad(explode(':', $host, 4), 4, null);
+
+            // set default values if not set
+            $host = ($host !== null) ? $host : '127.0.0.1';
+            $port = ($port !== null) ? $port : 6379;
+            $database = ($database !== null) ? $database : 0;
+
+            if ($this->redis->connect($host, $port) === false) {
+                rcube::raise_error(
+                    array(
+                        'code' => 604,
+                        'type' => 'session',
+                        'line' => __LINE__,
+                        'file' => __FILE__,
+                        'message' => "Could not connect to Redis server. Please check host and port"
+                    ),
+                    true,
+                    true
+                );
+            }
+
+            if ($password != null && $this->redis->auth($password) === false) {
+                rcube::raise_error(
+                    array(
+                        'code' => 604,
+                        'type' => 'session',
+                        'line' => __LINE__,
+                        'file' => __FILE__,
+                        'message' => "Could not authenticate with Redis server. Please check password."
+                    ),
+                    true,
+                    true
+                );
+            }
+
+            if ($database != 0 && $this->redis->select($database) === false) {
+                rcube::raise_error(
+                    array(
+                        'code' => 604,
+                        'type' => 'session',
+                        'line' => __LINE__,
+                        'file' => __FILE__,
+                        'message' => "Could not select Redis database. Please check database setting."
+                    ),
+                    true,
+                    true
+                );
+            }
+        }
+
+        // register sessions handler
+        $this->register_session_handler();
+    }
+
+    /**
+     * @param $save_path
+     * @param $session_name
+     * @return bool
+     */
+    public function open($save_path, $session_name)
+    {
+        return true;
+    }
+
+    /**
+     * @return bool
+     */
+    public function close()
+    {
+        return true;
+    }
+
+    /**
+     * remove data from store
+     *
+     * @param $key
+     * @return bool
+     */
+    public function destroy($key)
+    {
+        if ($key) {
+            $this->redis->del($key);
+        }
+
+        return true;
+    }
+
+
+    /**
+     * read data from redis store
+     *
+     * @param $key
+     * @return null
+     */
+    public function read($key)
+    {
+        if ($value = $this->redis->get($key)) {
+            $arr = unserialize($value);
+            $this->changed = $arr['changed'];
+            $this->ip      = $arr['ip'];
+            $this->vars    = $arr['vars'];
+            $this->key     = $key;
+
+            return !empty($this->vars) ? (string) $this->vars : '';
+        }
+
+        return null;
+    }
+
+
+    /**
+     * write data to redis store
+     *
+     * @param $key
+     * @param $newvars
+     * @param $oldvars
+     * @return bool
+     */
+    public function update($key, $newvars, $oldvars)
+    {
+        $ts = microtime(true);
+
+        if ($newvars !== $oldvars || $ts - $this->changed > $this->lifetime / 3) {
+            $this->redis->setex($key, $this->lifetime + 60, serialize(array('changed' => time(), 'ip' => $this->ip, 'vars' => $newvars)));
+        }
+
+        return true;
+    }
+
+
+    /**
+     * write data to redis store
+     *
+     * @param $key
+     * @param $vars
+     * @return bool
+     */
+    public function write($key, $vars)
+    {
+        return $this->redis->setex($key, $this->lifetime + 60, serialize(array('changed' => time(), 'ip' => $this->ip, 'vars' => $vars)));
+    }
+
+
+}
\ No newline at end of file

diff --git a/config/defaults.inc.php b/config/defaults.inc.php
index 50c392d..50ae71e 100644
--- a/config/defaults.inc.php
+++ b/config/defaults.inc.php
@@ -380,9 +380,14 @@
 // Session path. Defaults to PHP session.cookie_path setting.
 $config['session_path'] = null;
 
 // Backend to use for session storage. Can either be 'db' (default), 'memcache' or 'php'
 // If set to 'memcache', a list of servers need to be specified in 'memcache_hosts'
 // Make sure the Memcache extension (http://pecl.php.net/package/memcache) version >= 2.0.0 is installed
+//
+// If set to 'redis', a server needs to be specified in 'redis_hosts'
+// Make sure the Redis extension (http://pecl.php.net/package/redis) version >= 2.0.0 is installed
+//
 // Setting this value to 'php' will use the default session save handler configured in PHP
 $config['session_storage'] = 'db';
 
@@ -403,6 +408,13 @@
 // See http://php.net/manual/en/memcache.addserver.php
 $config['memcache_retry_interval'] = 15;
 
+// use this for accessing redis
+// currently only one host is supported. cluster support may come in a future release.
+// you can pass 4 fields, host, port, database and password.
+// unset fields will be set to the default values host=127.0.0.1, port=6379, database=0, password=  (empty)
+
+$config['redis_hosts'] = null; // e.g. array( 'localhost:6379' );  array( '192.168.1.1:6379:1:secret' );
+
 // check client IP in session authorization
 $config['ip_check'] = false;
 
@@ -380,9 +380,14 @@
 $config['session_path'] = null;
 
-// Backend to use for session storage. Can either be 'db' (default), 'memcache' or 'php'
+// Backend to use for session storage. Can either be 'db' (default), 'redis', 'memcache', or 'php'
+//
 // If set to 'memcache', a list of servers need to be specified in 'memcache_hosts'
 // Make sure the Memcache extension (http://pecl.php.net/package/memcache) version >= 2.0.0 is installed
--
Gitblit v1.9.1